/*
 *  https.c -- connect procedure for https & socks4/5 proxy
 *
 *  https.c is an addition part of binkd project
 *
 *  Copyright (C) 1996-1998  Dima Maloff, 5047/13
 *  Copyright (C) 1998-2000  Dima Afanasiev, 5020/463
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. See COPYING.
 */
/*
 * $Id: https.c,v 2.3 2001/09/14 07:24:20 gul Exp $
 *
 * $Log: https.c,v $
 * Revision 2.3  2001/09/14 07:24:20  gul
 * bindaddr works on outgoing connections now
 *
 * Revision 2.2  2001/07/28 17:26:26  gul
 * Avoid compiler warnings
 *
 * Revision 2.1  2001/06/19 15:49:13  da
 * *** empty log message ***
 *
 * Revision 2.0  2001/01/10 12:12:38  gul
 * Binkd is under CVS again
 *
 *
 */

#ifdef HTTPS

#include "binkd.h"
#ifdef NTLM
#include "ntlm/helpers.h"
#endif

/*
 * Do a proxy handshake depending of connect_method
 * Return 0 on success, -1 on error
 */

enum {
  CSTATE_SENDREQ,      /* send initial request to proxy                */
  CSTATE_WAITAUTH1,    /* wait for possible auth types (socks5)        */
  CSTATE_WAITAUTH2,    /* wait for "username ok" response (socks5)     */
  CSTATE_SENDCONNECT,  /* send destination address here (socks5)       */
  CSTATE_WAITCONNECTPART1, /* wait for fixed part of connect resp (socks5) */
  CSTATE_WAITCONNECT,  /* wait for final "connected ok" response       */
  CSTATE_CONNECTED,    /* connected                                    */
  CSTATE_DEAD,         /* error                                        */
  CSTATE_READHTTPLINE, /* special state for HTTP substates             */
#ifdef NTLM
  CSTATE_SKIPBYTES,    /* safely skip "content-length" bytes           */
#endif
};

enum
{
  HTTP_READCODE,     /* read first line with numeric response        */
  HTTP_WAITEND,      /* skip until end of header and report success  */
  HTTP_DEADEND,      /* skip until end of header and report error    */
#ifdef NTLM
  HTTP_WAITNTLMAUTH, /* wait for NTLM auth response before pass 2    */
#endif
};

static void dump_buffer(char *p, int size)
{
  char buf[100];
  int  i, n;

  while (size)
  {
    n = sprintf(buf, "Dump:");
    for (i = 0; i < 16 && size; p++, i++, size--)
      n += sprintf(buf+n, " %02X", *p);
    Log(7, "%s", buf);
  }
}

static int makecbstring(char *buf, char *string, int startpos)
{
  int n = strlen(string);

  if (n > 255)
    n = 255;
  buf[startpos++] = n;
  if (n)
    memcpy(buf+startpos, string, n);
  return startpos + n;
}

static char b64t[]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
int enbase64(char *data, int size, char *p)
{
  int i;
  int c;
  unsigned char *q;
  char *s = p;
  q = (unsigned char*)data;
  i=0;
  for(i = 0; i < size;){
    c=q[i++];
    c*=256;
    if(i < size)
      c+=q[i];
    i++;
    c*=256;
    if(i < size)
      c+=q[i];
    i++;
    p[0]=b64t[(c&0x00fc0000) >> 18];
    p[1]=b64t[(c&0x0003f000) >> 12];
    p[2]=b64t[(c&0x00000fc0) >> 6];
    p[3]=b64t[(c&0x0000003f) >> 0];
    if(i > size)
      p[3]='=';
    if(i > size+1)
      p[2]='=';
    p+=4;
  }
  return p - s;
}

#ifdef NTLM
int debase64(char *s, unsigned char *rc)
{
  int i, j;
  char *sp;
  if ((!s) || (!rc)) return (-1);
  for(i = j = 0; ((sp = strchr(b64t, s[i])) != NULL); i++)
  {
    int k = (sp - b64t);
    switch(i%4)
    {
    case 0: rc[j] = k<<2; break;
    case 1: rc[j++] |= k>>4; rc[j] = (k << 4) & 0xF0; break;
    case 2: rc[j++] |= k>>2; rc[j] = (k << 6) & 0xC0; break;
    case 3: rc[j++] |= k; break;
    }
  }
  return j;
}
#endif

char connect_method_name[][7] = { "dir", "socks4", "socks5", "http" };

int net_handshake(int sockfd, int connect_method,
                  char *node_host, int node_port,
                  char *proxy_user, char *proxy_password
                 )
{
  int  ntlm = 0;
  int  state, prevstate, httpstate;
  int  toberead = 0;
  int  tobewritten = 0;
  int  readto_pos = 0;
  char databuf[1000]; /* socks username+password 255 each max or http request */
  int  n;
#ifdef NTLM
  char *ntlm_host, *ntlm_domain;
  char *ntlm_saved_error = NULL;
  char *ntlm_answer      = NULL;
  int   ntlm_content_length = 0;
#endif

  if (connect_method == CMETHOD_DIRECT) /* do nothing */
    return 0;

  Log(4, "trying %s:%d thru %s proxy...", node_host, node_port,
      connect_method_name[connect_method]);

  for (prevstate = state = CSTATE_SENDREQ;
       state != CSTATE_DEAD && state != CSTATE_CONNECTED;
      )
  {
    if (tobewritten)
    {
      Log(6, "proxy: %d bytes to send", tobewritten);

      dump_buffer(databuf, tobewritten);

      n = send(sockfd, databuf, tobewritten, 0);
      if (n != tobewritten)
      {
        if (n < 0)
          Log(1, "Proxy write error: %s", TCPERR());
        else
          Log(1, "Proxy write error: only %d of %d bytes sent",
              n, tobewritten);
        state = CSTATE_DEAD;
      }
      tobewritten = 0;
    }
    else if (toberead)
    {
      n = recv(sockfd, databuf + readto_pos, toberead, 0);
      if (n > 0)
      {
        Log(6, "proxy: read %d bytes this time", n);
        dump_buffer(databuf + readto_pos, n);
        toberead   -= n;
        readto_pos += n;
      }
      else
      {
        if (n < 0)
          Log(1, "Proxy read error: %s", TCPERR());
        else
          Log(1, "Proxy read: Connection closed by proxy");
        state = CSTATE_DEAD;
      }
    }

    if (tobewritten != 0 || toberead != 0)
      continue; /* do not run state machine if socket rw pending */

    switch (connect_method)
    {
      /*************** SOCKS4 states **************/
    case CMETHOD_SOCKS4:

      if (state == CSTATE_SENDREQ)
      {
        struct resolve_info res;
        struct hostent *hp;

        /* socks4 want real IP of peer to connect to */
        hp = net_resolve(node_host, &res);
        if (hp == NULL)
        {
          state = CSTATE_DEAD;
          break;
        }

        databuf[0] = 4;
        databuf[1] = 1;
        *(unsigned short *)(databuf+2) = htons(node_port);
        memcpy(databuf+4, hp->h_addr_list[0], 4);
        databuf[8] = 0;

        net_freeresolve(&res);

        tobewritten = 9;
        toberead    = 8;
        state       = CSTATE_WAITCONNECT;
      } else if (state == CSTATE_WAITCONNECT)
      {
        /* assume error by default */
        state = CSTATE_DEAD;
        /* process server response */
        if (databuf[0] != 0)
        {
          Log(1, "socks4: bad version in response (%d)", databuf[0]);
          break;
        }
        switch (databuf[1])
        {
        case 90: state = CSTATE_CONNECTED; break; /* the only success way */
        case 91: Log(1, "socks4: connection rejected by server");   break;
        case 92: Log(1, "socks4: server said `ident failed'");   break;
        case 93: Log(1, "socks4: server said `ident conflict'"); break;
        default: Log(1, "socks4: server said `error %d'", databuf[1]); break;
        }
      } /* if state */

      break;
      /*************** SOCKS5 states *****************/

    case CMETHOD_SOCKS5:
      if (state == CSTATE_SENDREQ)
      {
        databuf[0] = 5;  /* socks version */
        databuf[2] = 0;  /* always suggest auth method 0 */
        if (*proxy_user == 0)
        {
          databuf[1]  = 1;  /* use only method 0 */
          tobewritten = 3;
        }
        else
        {
          databuf[1]  = 2;  /* total 2 methods  */
          databuf[3]  = 2;  /* use method 2     */
          tobewritten = 4;
        }

        toberead = 2;
        state    = CSTATE_WAITAUTH1;
      } else if (state == CSTATE_WAITAUTH1)
      {
        state = CSTATE_DEAD; /* assume error by default */
        if (databuf[0] != 5)
        {
          Log(1, "socks5: wrong version in response (%d)", databuf[0]);
          break;
        }
        switch (databuf[1])
        {
        default: Log(1, "socks5: authorization type rejected by server"); break;
        case 0:  state = CSTATE_SENDCONNECT; break; /* no auth */
        case 2:
          databuf[0] = 1;  /* version */
          tobewritten = makecbstring(databuf, proxy_user, 1);
          tobewritten = makecbstring(databuf, proxy_password, tobewritten);

          toberead    = 2;
          state       = CSTATE_WAITAUTH2;
          break;
        }
      } else if (state == CSTATE_WAITAUTH2)
      {
        state = CSTATE_DEAD; /* assume error by default */
        if (databuf[0] != 1)
        {
          Log(1, "socks5: wrong version in response (%d)", databuf[0]);
          break;
        }
        if (databuf[1] != 0)
        {
          Log(1, "socks5: server rejected username/password (code %d)", databuf[1]);
          break;
        }
        state = CSTATE_SENDCONNECT;
      } else if (state == CSTATE_SENDCONNECT)
      {
        databuf[0] = 5;  /* version           */
        databuf[1] = 1;  /* command: connect  */
        databuf[2] = 0;  /* reserved          */
        databuf[3] = 3;  /* address type: domain name */
        tobewritten = makecbstring(databuf, node_host, 4);
        *(unsigned short *)(databuf+tobewritten) = htons(node_port);
        tobewritten += 2;
        toberead    = 5;  /* get fixed part and 1st byte of address */
        state       = CSTATE_WAITCONNECTPART1;
      } else if (state == CSTATE_WAITCONNECTPART1)
      {
        char *err;
        state = CSTATE_DEAD;
        if (databuf[0] != 5)
        {
          Log(1, "socks5: wrong version in response (%d)", databuf[0]);
          break;
        }
        switch (databuf[1])
        {
        case 0:  err = NULL; break;
        case 1:  err = "general socks server failure"; break;
        case 2:  err = "connection not allowed by ruleset"; break;
        case 3:  err = "network unreachable"; break;
        case 4:  err = "host unreachable"; break;
        case 5:  err = "connection refused"; break;
        case 6:  err = "TTL expired"; break;
        case 7:  err = "command not supported"; break;
        case 8:  err = "address type not supported"; break;
        default: err = "unknown error"; break;
        }
        if (err)
        {
          Log(1, "socks5: server said '%s' (code %d)", err, databuf[1]);
          break;
        }
        /* length of address part (full) */
        switch (databuf[3])
        {
        case 1:  toberead = 4; break;            /* 4 bytes of IP v4  */
        case 3:  toberead = databuf[4]+1; break; /* var.length name   */
        case 4:  toberead = 16; break;           /* 16 bytes of IP v6 */
        default: toberead = 0;  break;
        }
        if (toberead == 0)
        {
          Log(1, "socks5: unknown address type in response (%d)", databuf[3]);
          break;
        }
        /* eat name (we already eat one byte!) and port */
        toberead += (2 - 1);
        state = CSTATE_WAITCONNECT;
      } else if (state == CSTATE_WAITCONNECT)
      {
        /* rest of IP address purged Ok */
        state = CSTATE_CONNECTED;
      }
      break;

    case CMETHOD_HTTP:
      if (state == CSTATE_SENDREQ)
      {
#ifdef NTLM
        if (ntlm == 0)  /* detect presense of ntlm */
        {
          char *p;

          ntlm_host = ntlm_domain = "";
          if ((p = strchr(proxy_password, '/')) != NULL)
          {
            ntlm = 1;
            *p++ = 0;
            ntlm_host = p;
            if ((p = strchr(ntlm_host, '/')) != NULL)
            {
              *p++ = 0;
              ntlm_domain = p;
            }
            /* Log(1, "nltm host='%s', domain='%s'", ntlm_host, ntlm_domain); */
          }
        }
#endif
        tobewritten = sprintf(databuf, "CONNECT %s:%d HTTP/1.%d\r\n",
                              node_host, node_port, (ntlm != 0));
        if (*proxy_user)
        {
#ifdef NTLM
          if (ntlm)
          {
            if (ntlm == 1)
              tobewritten += sprintf(databuf+tobewritten, "Connection: keep-alive\r\n");
            tobewritten += sprintf(databuf+tobewritten, "Proxy-Authorization: NTLM ");
            if (ntlm == 1)
              tobewritten += getNTLM1(databuf+tobewritten, ntlm_host, ntlm_domain);
            else
              tobewritten += getNTLM2(databuf+tobewritten, ntlm_answer, proxy_user, proxy_password, ntlm_host, ntlm_domain);
            databuf[tobewritten++] = '\r';
            databuf[tobewritten++] = '\n';
          }
          else
#endif
          {
            char *sp, *sp2;

            n = strlen(proxy_user) + strlen(proxy_password) + 2;
            sp  = xalloc(n);
            sp2 = xalloc(n * 4 / 3 + 4);
            sprintf(sp, "%s:%s", proxy_user, proxy_password);
            sp2[enbase64(sp, n-1, sp2)] = 0;
            tobewritten += sprintf(databuf+tobewritten, "Proxy-Authorization: Basic %s\r\n", sp2);
            free(sp2);
            free(sp);
          }
        }
        databuf[tobewritten++] = '\r';
        databuf[tobewritten++] = '\n';

        toberead  = 1;
        state     = CSTATE_READHTTPLINE;
        httpstate = HTTP_READCODE;
      } else if (state == CSTATE_READHTTPLINE)
      {
        if (readto_pos < 1 || databuf[readto_pos-1] != '\n')
        {
          /* read once again */
          if (readto_pos > sizeof(databuf)-2)
          {
            /* protect over buffer overflow - ignore long responses */
            readto_pos--;
            databuf[readto_pos-1] = databuf[readto_pos];
          }
          toberead = 1;
          break;
        }

        /* Kill '\n' and optional '\r' */
        databuf[--readto_pos] = 0;
        if (readto_pos && databuf[readto_pos-1] == '\r')
          databuf[readto_pos-1] = 0;
        readto_pos = 0;

        Log(6, "HTTP response: '%s'", databuf);

        if (*databuf == 0)
        {
          if (httpstate == HTTP_READCODE)
          {
            /* empty lines before response must be ingnored
             * according to rfc-2616 */
            toberead   = 1; /* just continue reading new line */
          } else if (httpstate == HTTP_WAITEND)
          {
            state = CSTATE_CONNECTED;
          } else if (httpstate == HTTP_DEADEND)
          {
            state = CSTATE_DEAD;
#ifdef NTLM
          } else if (httpstate == HTTP_WAITNTLMAUTH)
          {
            if (ntlm != 2) /* handshake not suggested */
            {
              Log(1, "NTLM handshake not requested, proxy said '%s'", ntlm_saved_error);
              state = CSTATE_DEAD;
            }
            else
            {
              state    = CSTATE_SKIPBYTES;  /* skip ntlm_content_length answer */
            }
#endif
          } else
          {
            Log(1, "Unexpected end of http response");
            state = CSTATE_DEAD;
          }
          break;
        }

        /* got a non-empty line */
        if (httpstate == HTTP_READCODE)
        {
          char *s;

          if (memcmp(databuf, "HTTP", 4) != 0 || (s = strchr(databuf, ' ')) == NULL)
          {
            Log(1, "Wrong proxy response `%s'", databuf);
            state = CSTATE_DEAD;
            break;
          }
          while (isspace(*s))
            s++;

          /* 200 - Ok, 407 - auth required */

          if (atoi(s) == 200)
          {
            /* success - skip rest of the headers */
            httpstate = HTTP_WAITEND;
          }
          else
          {
#ifdef NTLM
            if (ntlm == 1)
            {
              /* keep error message */
              ntlm_saved_error = xstrdup(s);
              /* NTLM pass 1 - wait for NTLM auth response */
              httpstate = HTTP_WAITNTLMAUTH;
            }
            else
#endif
            {
              Log(1, "http proxy reported error: '%s'", s);
              httpstate = HTTP_DEADEND;
              /* we're not aborting here, user can see more
               * headers in log file for further explanation */
            }
          }
        }
#ifdef NTLM
        else if (httpstate == HTTP_WAITNTLMAUTH)
        {
          static char spaces[] = " \t";
          char *word1, *word2;

          if ((word1 = strtok(databuf, spaces)) != NULL &&
              (word2 = strtok(NULL, spaces)) != NULL)
          {
            /* Authenticate: NTLM bla-bla-bla */
            if (strstr(word1, "uthenticate:") != NULL &&
                STRICMP(word2, "NTLM") == 0)
            {
              if ((word1 = strtok(NULL, spaces)) != NULL)
              {
                ntlm = 2;  /* moving to ntlm state 2 */
                ntlm_answer = xstrdup(word1);
              }
            }
            /* Content-Length: bla-bla-bla */
            if (STRICMP(word1, "Content-Length:") == 0)
              ntlm_content_length = atol(word2);
          }
        }
#endif
        state     = CSTATE_READHTTPLINE;
        toberead  = 1; /* continue reading */
      }
#ifdef NTLM
      else if (state == CSTATE_SKIPBYTES)
      {
        /* skip long response (more than databuf size) */
        if (ntlm_content_length == 0)
          state = CSTATE_SENDREQ;
        else
        {
          readto_pos = 0;
          toberead   = ntlm_content_length > sizeof(databuf) ?
            sizeof(databuf) : ntlm_content_length;
          ntlm_content_length -= toberead;
        }
      }
#endif
      break;
    } /* switch (connect_method) */

    if (state != prevstate)
    {
      Log(6, "proxy: state changed %d => %d", prevstate, state);
      prevstate  = state;
      readto_pos = 0;
    }
  }

#ifdef NTLM
  xfree(ntlm_saved_error);
  xfree(ntlm_answer);
#endif

  return state == CSTATE_CONNECTED ? 0 : -1;
}
#endif
